#version 330
#extension GL_ARB_separate_shader_objects : enable





layout(location = 0) smooth in vec2 _fin_fragTexCoord;

struct SFragmentInput
{
	vec2 fragTexCoord;
	bool IsFrontFace;
};

layout(location = 0) out vec4 _fout_outColor;

struct SFragmentOutput
{
	vec4 outColor;
};

layout( std140) uniform USceneInfo
{
	mat4 ViewProj;
	mat4 View;
	mat4 Proj;
	mat4 InvView;
	mat4 InvViewProj;
	mat4 UserToLHZ;
	mat4 LHZToUser;
	mat4 BillboardingView;
	mat4 PackNormalView;
	mat4 UnpackNormalView;
	vec4 SideVector;
	vec4 DepthVector;
	vec2 ZBufferLimits;
	float Handedness;
} SceneInfo;
layout( std140) uniform ULightInfo
{
	vec4 LightParam0;
	vec4 LightParam1;
	vec3 Color;
	vec3 Ambient;
} LightInfo;
 uniform sampler2D BRDFLUTSampler;
 uniform samplerCube EnvironmentMapSampler;
 uniform sampler2D DepthSampler;
 uniform sampler2D NormalRoughMetalSampler;
 uniform sampler2D DiffuseSampler;



vec3 DepthToWorldPos(float pDepthValue, vec2 pUv, mat4 pInvViewProj)
{
	vec2 clipSpace = pUv * 2.0 - 1.0;
	vec4 projPos = vec4(clipSpace, pDepthValue, 1.0);
	projPos =  ((pInvViewProj) * (projPos)) ;
	return projPos.xyz / projPos.w;
}

vec3 UnpackNormalSpheremap(vec2 normal  )
{
	vec2 fenc = normal * 4.0f - 2.0f;
	float f = dot(fenc, fenc);
	float g = sqrt(1.0f - f / 4.0f);
	vec3 outNormal;
	outNormal.xy = fenc * g;
	outNormal.z = 1.0f - f / 2.0f;
	return  (( SceneInfo.UnpackNormalView ) * (vec4(outNormal, 0.0f))) .xyz;
}






vec3 fresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness)
{
	return F0 + (max(vec3(1.0f - roughness, 1.0f - roughness, 1.0f - roughness), F0) - F0) * pow(1.0f - cosTheta, 5.0f);
}

vec3 computeAmbientBRDF(vec3 surfToView, vec3 surfaceNormal, float roughness, float metalness, vec3 surfaceColor  )
{

	const float kCubemapMipRange = 6;

	vec3 surfaceMetalColor = mix( vec3(0.04, 0.04, 0.04) , surfaceColor, metalness);
	vec3 kS = fresnelSchlickRoughness(max(dot(surfaceNormal, surfToView), 0.0), surfaceMetalColor, roughness);
	vec3 kD = 1.0f - kS;
	kD *= 1.0f - metalness;


	vec3 irradiance =  textureLod(EnvironmentMapSampler, surfaceNormal, kCubemapMipRange) .rgb;
	vec3 diffuse = irradiance * surfaceColor;


	vec3 sampleDir = reflect(-surfToView, surfaceNormal);
	vec3 LHZSampleDir =  (( SceneInfo.UserToLHZ ) * (vec4(sampleDir, 0.0f))) .xyz;
	vec3 RHYSampleDir = vec3(LHZSampleDir.x, LHZSampleDir.z, LHZSampleDir.y);
	vec3 prefilteredColor =  textureLod(EnvironmentMapSampler, RHYSampleDir, pow(roughness, 0.5f) * kCubemapMipRange) .rgb;
	vec2 envBRDF =  texture(BRDFLUTSampler, vec2(max(dot(surfaceNormal, surfToView), 0.0), roughness)) .rg;
	vec3 specular = prefilteredColor * (kS * envBRDF.x + envBRDF.y);

	vec3 ambient = kD * diffuse + specular;

	return ambient;
}

void FragmentMain( in SFragmentInput  fInput,  out SFragmentOutput  fOutput  )
{
	vec3 ambientColor =  LightInfo.Ambient ;
	mat4 invViewProj =  SceneInfo.InvViewProj ;
	vec3 viewPosition =  SceneInfo.UnpackNormalView [3] .xyz;

	vec3 surfaceColor =  texture(DiffuseSampler, fInput.fragTexCoord) .rgb;
	float depthValue =  texture(DepthSampler, fInput.fragTexCoord) .r;
	vec3 surfacePosition = DepthToWorldPos(depthValue, fInput.fragTexCoord, invViewProj);
	vec4 packedNormalRoughMetal =  texture(NormalRoughMetalSampler, fInput.fragTexCoord) ;
	vec3 surfaceNormal = UnpackNormalSpheremap(packedNormalRoughMetal.xy  );
	float surfaceRoughness = packedNormalRoughMetal.z;
	float surfaceMetalness = packedNormalRoughMetal.w;

	if (surfaceRoughness < 0.0f && surfaceMetalness < 0.0f)
	{

		fOutput.outColor = vec4(surfaceColor, 0.0f);
		return;
	}

	surfaceRoughness = max(0.01f, surfaceRoughness);
	surfaceMetalness = max(0.01f, surfaceMetalness);

	vec3 surfToView = viewPosition - surfacePosition;
	vec3 surfToViewDir = normalize(surfToView);

	vec3 lightAmbient = computeAmbientBRDF( surfToViewDir,
												surfaceNormal,
												surfaceRoughness,
												surfaceMetalness,
												surfaceColor
												);
	fOutput.outColor = vec4(lightAmbient * ambientColor, 0.0f);
}


void main()
{
	SFragmentInput fInput;
	SFragmentOutput fOutput;

	fInput.fragTexCoord = _fin_fragTexCoord;
	fInput.IsFrontFace = gl_FrontFacing;
	FragmentMain(fInput, fOutput);
	_fout_outColor = fOutput.outColor;
}
